"""Implementation of XBAreaction class.

Peter Schubert, HHU Duesseldorf, May 2022
"""

import re
import sbmlxdf
from xbanalysis.utils.utils import get_species_stoic, get_local_params, expand_kineticlaw


class XbaReaction:

    def __init__(self, s_reaction, functions):
        self.id = s_reaction.name
        self.name = s_reaction.get('name', '')
        self.sboterm = s_reaction['sboterm']
        self.reaction_string = s_reaction['reactionString']
        self.reversible = s_reaction['reversible']
        self.products = get_species_stoic(s_reaction['products'])
        self.reactants = get_species_stoic(s_reaction['reactants'])
        self.modifiers = get_species_stoic(s_reaction['modifiers'])
        self.local_params = get_local_params(s_reaction)
        self.kinetic_law = s_reaction['kineticLaw'] if type(s_reaction['kineticLaw']) == str else ''
        # convert to numpy, replace local parameters with numerical values, inline lambda functions:
        expanded_kl = sbmlxdf.misc.mathml2numpy(self.kinetic_law)
        for pid, val in self.local_params.items():
            expanded_kl = re.sub(r'\b' + pid + r'\b', str(val), expanded_kl)
        self.expanded_kl = expand_kineticlaw(expanded_kl, functions)
        self.ps_product = ''
        self.enzyme = ''

    def set_enzyme(self, species):
        for sid in self.modifiers:
            if species[sid].sboterm == 'SBO:0000250' or species[sid].sboterm == 'SBO:0000252':
                self.enzyme = sid
                break

    def add_species_usage(self, species):
        # for metabolic reactions add to the species consumed/produced the metabolic reaction id
        # !!! we could also enter the Reaction Object references instead !!!!!!!!!!
        if (self.sboterm == 'SBO:0000167') or (self.sboterm == 'SBO:0000179') or (self.sboterm == 'SBO:0000182'):
            for sid, val in self.products.items():
                species[sid].add_m_reaction(self.id, val)
            for sid, val in self.reactants.items():
                species[sid].add_m_reaction(self.id, -val)
        # protein degradation reaction, assume that first protein in reactants is getting degraded
        # a protein degradation reaction is still a metabolic reaction
        if self.sboterm == 'SBO:0000179':
            for sid in self.reactants:
                if (species[sid].sboterm == 'SBO:0000252') or (species[sid].sboterm == 'SBO:0000250'):
                    species[sid].set_p_degrad_rid(self.id)
                    break
        # protein interconversion reactions
        if self.sboterm == 'SBO:0000182':
            for sid in self.reactants:
                if (species[sid].sboterm == 'SBO:0000252') or (species[sid].sboterm == 'SBO:0000250'):
                    species[sid].set_p_conversion_rid(self.id)
                    break
            for sid in self.products:
                if (species[sid].sboterm == 'SBO:0000252') or (species[sid].sboterm == 'SBO:0000250'):
                    species[sid].set_p_conversion_rid(self.id)
                    break

        # for protein synthesis reactions identify the enzyme being produced from reaction.products
        #     to species being consumed add the enzyme product with corresponding stoichiometry
        #     to species being produced add the enzyme product with corresponding stoichiometry, unless
        #        the species is itself the enzyme product, in which case we add the ps reaction producing it.
        if self.sboterm == 'SBO:0000589':
            enz_sid = ''
            for sid in self.products:
                if (species[sid].sboterm == 'SBO:0000252') or (species[sid].sboterm == 'SBO:0000250'):
                    enz_sid = sid
                    self.ps_product = enz_sid
                    species[enz_sid].set_ps_rid(self.id)
                    break
            for sid, val in self.products.items():
                if sid != enz_sid:
                    species[sid].add_ps_reaction(enz_sid, val)
            for sid, val in self.reactants.items():
                species[sid].add_ps_reaction(enz_sid, -val)
