"""Implementation of XBAmodel class.

Peter Schubert, HHU Duesseldorf, May 2022
"""

import re
import numpy as np
import sympy
import types


class GbaProblem:

    def __init__(self, xba_model):
        self.xba_model = xba_model

        # metabolites and enzymes in the model
        metab_sids = [s.id for s in xba_model.species.values() if s.sboterm == 'SBO:0000247']
        enz_sids = [s.id for s in xba_model.species.values() if (s.sboterm == 'SBO:0000250' or
                                                                 s.sboterm == 'SBO:0000252')]

        # optimization variables
        self.var_metab_sids = [sid for sid in metab_sids if xba_model.species[sid].constant is False]
        self.var_enz_sids = [sid for sid in enz_sids if xba_model.species[sid].constant is False]
        self.var_sids = self.var_metab_sids + self.var_enz_sids
        self.var_n_metabs = len(self.var_metab_sids)
        self.var_sid2idx = {sid: idx for idx, sid in enumerate(self.var_sids)}
        self.var_initial = np.array([xba_model.species[sid].initial_conc for sid in self.var_sids])
        #        self.var_lower_bounds = np.zeros(len(self.var_sids)) * 1
        #        self.var_upper_bounds = np.ones(len(self.var_sids)) * 0.1

        # constant concentrations
        self.const_metab_sids = [sid for sid in metab_sids if xba_model.species[sid].constant is True]
        self.const_sid2idx = {sid: idx for idx, sid in enumerate(self.const_metab_sids)}
        self.const_enz_sids = [sid for sid in enz_sids if xba_model.species[sid].constant is True]

        self.all_enz_sids = self.var_enz_sids + self.const_enz_sids

        # reactions - yet to implement, accept child SBO terms, e.g. here of 00167
        self.mr_rids = [r.id for r in xba_model.reactions.values()
                        if ((r.sboterm == 'SBO:0000167') or
                            (r.sboterm == 'SBO:0000179') or
                            (r.sboterm == 'SBO:0000182'))]
        self.ps_rids = [r.id for r in xba_model.reactions.values() if r.sboterm == 'SBO:0000589']
        spont_rids = [rid for rid in self.mr_rids if xba_model.reactions[rid].enzyme == '']

        # calculate stoichiometric for mass balance equality constraints
        s_mat = self.get_stoic_matrix(self.var_metab_sids, self.mr_rids)
        var_sids_vol = np.array([self.xba_model.compartments[self.xba_model.species[sid].compartment].size
                                 for sid in self.var_metab_sids])
        s_scaled_mat = s_mat/var_sids_vol.reshape(-1, 1)

        all_enz_ps_rids = [xba_model.species[sid].ps_rid for sid in self.all_enz_sids]
        p_mat = self.get_stoic_matrix(self.var_metab_sids, all_enz_ps_rids)
        all_enz_vol = np.array([self.xba_model.compartments[self.xba_model.species[sid].compartment].size
                                for sid in self.all_enz_sids])
        vol_scale_p_mat = np.outer(1.0/var_sids_vol, all_enz_vol)
        p_scaled_mat = p_mat * vol_scale_p_mat

        # model parameters
        self.model_params = {pid: p.value for pid, p in xba_model.parameters.items()}
        self.model_params.update({cid: c.size for cid, c in xba_model.compartments.items()})
        self.model_params['macro'] = self.model_params.get('macro', 0.0) > 0.0
        self.model_params['ext_conc'] = np.array([xba_model.species[sid].initial_conc
                                                  for sid in self.const_metab_sids])
        self.model_params['const_enz_conc'] = np.array([xba_model.species[sid].initial_conc
                                                        for sid in self.const_enz_sids])
        self.model_params['S_scaled_mat'] = s_scaled_mat
        self.model_params['P_scaled_mat'] = p_scaled_mat
        # self.model_params['P_scaled_csr'] = scipy.sparse.csr_matrix(p_mat * vol_scale_p_mat)
        self.model_params['np'] = np
        self.model_params['weights'] = np.array([xba_model.species[sid].weight for sid in self.var_sids])
        self.model_params['const_enz_weights'] = np.array([xba_model.species[sid].weight
                                                           for sid in self.const_enz_sids])
        self.model_params['weights'] = np.array([xba_model.species[sid].weight for sid in self.var_sids])

        # create functions
        self.mrs, self.mrs_jac = self.get_reactions(self.mr_rids)
        self.spontrs, self.spontrs_jac = self.get_reactions(spont_rids)
        self.pss, self.pss_jac = self.get_reactions(self.ps_rids)
        self.inv_ps_rates, self.inv_ps_rates_jac = self.get_inv_ps_rates()
        #        self.gammas, self.gammas_jac = self.get_gammas()
        self.p_degrads, self.p_degrads_jac = self.get_p_degrad_rates()
        self.enz_conv_mb, self.enz_conv_mb_jac = self.get_enz_interconv_mb()

    def get_stoic_matrix(self, sids, rids):
        # matrix for free metabolites and list of reactions
        stoic_mat = np.zeros((len(sids), len(rids)))
        map_metab_row = {sid: idx for idx, sid in enumerate(sids)}
        for col, rid in enumerate(rids):
            for sid, stoic in self.xba_model.reactions[rid].reactants.items():
                if sid in map_metab_row:
                    stoic_mat[map_metab_row[sid], col] = -stoic
            for sid, stoic in self.xba_model.reactions[rid].products.items():
                if sid in map_metab_row:
                    stoic_mat[map_metab_row[sid], col] = stoic
        return stoic_mat

    def update_model_params(self, key, val):
        self.model_params[key] = val

    def _to_vectors(self, rs_str):
        # replace free and constant metabolites by vector elements
        rs_v_str = []
        for r_v_str in rs_str:
            for idx, sid in enumerate(self.var_sids):
                r_v_str = re.sub(r'\b' + sid + r'\b', f'x[{idx}]', r_v_str)
            for idx, sid in enumerate(self.const_metab_sids):
                r_v_str = re.sub(r'\b' + sid + r'\b', f'ext_conc[{idx}]', r_v_str)
            for idx, sid in enumerate(self.const_enz_sids):
                r_v_str = re.sub(r'\b' + sid + r'\b', f'const_enz_conc[{idx}]', r_v_str)
            rs_v_str.append(r_v_str)
        return rs_v_str

    def _gradients(self, rs_v_str):
        # generate derivatives
        x = sympy.IndexedBase('x')
        ext_conc = sympy.IndexedBase('ext_conc')
        const_enz_conc = sympy.IndexedBase('const_enz_conc')
        grads_str = []
        for i, r_v_str in enumerate(rs_v_str):
            func_sym = sympy.sympify(r_v_str, locals={'x': x,
                                                      'ext_conc': ext_conc,
                                                      'const_enz_conc': const_enz_conc})
            partial_der = [str(func_sym.diff(x[idx])) for idx in range(len(self.var_sids))]
            grads_str.append('[' + ', '.join(partial_der) + ']')
        #            progress_bar(i, len(rs_v_str))
        jac = compile('def jac(x): return np.array([' + ', '.join(grads_str) + '])',
                      '<string>', 'exec')
        return jac

    def get_reactions(self, rids):
        kinetics_str = [self.xba_model.reactions[rid].expanded_kl for rid in rids]
        v_str = self._to_vectors(kinetics_str)
        func_code = compile('def reactions(x): return np.array([' + ', '.join(v_str) + '])',
                            '<string>', 'exec')
        rs = types.FunctionType(func_code.co_consts[0], self.model_params)
        rs_jac = types.FunctionType(self._gradients(v_str).co_consts[0], self.model_params)
        return rs, rs_jac

    def get_inv_ps_rates(self):
        # inverse of max protein synthesis rates as function of variable enzymes and constant enzymes
        ps_inv_rs_str = []
        for sid in self.all_enz_sids:
            sp = self.xba_model.species[sid]
            ps_rid = sp.ps_rid
            ps_inv_rs_str.append(f'{sp.compartment}/({self.xba_model.reactions[ps_rid].expanded_kl})')
        v_str = self._to_vectors(ps_inv_rs_str)
        func_code = compile('def ps_inv(x): return np.array([' + ', '.join(v_str) + '])',
                            '<string>', 'exec')
        inv_ps_rates = types.FunctionType(func_code.co_consts[0], self.model_params)
        inv_ps_rates_jac = types.FunctionType(self._gradients(v_str).co_consts[0], self.model_params)
        return inv_ps_rates, inv_ps_rates_jac

    def get_p_degrad_rates(self):
        # protein degradation rates
        degrads_str = []
        for sid in self.all_enz_sids:
            sp = self.xba_model.species[sid]
            if sp.p_degrad_rid is not None:
                degrads_str.append(f'({self.xba_model.reactions[sp.p_degrad_rid].expanded_kl})/' +
                                   f'{sp.compartment}')
            else:
                degrads_str.append('0')
        v_str = self._to_vectors(degrads_str)
        func_code = compile('def p_degrads(x): return np.array([' + ', '.join(v_str) + '])',
                            '<string>', 'exec')
        degrads = types.FunctionType(func_code.co_consts[0], self.model_params)
        degrads_jac = types.FunctionType(self._gradients(v_str).co_consts[0], self.model_params)
        return degrads, degrads_jac

    def get_enz_interconv_mb(self):
        # mass balance for interconvering enzymes
        # identify enzymes in ExMr submatrix (enzymes x metabolic reactions)
        #  identify sets of enzymes that interconvert (consider unique sets)
        enz_in_mrs_sids = []
        for enz_sid in self.all_enz_sids:
            sp = self.xba_model.species[enz_sid]
            partner_enz_sids = {enz_sid}
            for mr_rid, enz_stoic in sp.m_reactions.items():
                rid = self.xba_model.reactions[mr_rid]
                if enz_stoic < 0:
                    for other_sid, other_stoic in rid.products.items():
                        sp_other = self.xba_model.species[other_sid]
                        if (sp_other.sboterm == 'SBO:0000250') or (sp_other.sboterm == 'SBO:0000252'):
                            partner_enz_sids.add(other_sid)
                if enz_stoic > 0:
                    for other_sid, other_stoic in rid.reactants.items():
                        sp_other = self.xba_model.species[other_sid]
                        if (sp_other.sboterm == 'SBO:0000250') or (sp_other.sboterm == 'SBO:0000252'):
                            partner_enz_sids.add(other_sid)
            # look for enzyme interconversion
            if len(partner_enz_sids) > 1:
                # consider unique interconversion sets
                if partner_enz_sids not in enz_in_mrs_sids:
                    enz_in_mrs_sids.append(partner_enz_sids)

        # for each set interconverting enzymes create a mass balance equation
        conversions_str = []
        for partner_enz_sids in enz_in_mrs_sids:
            enz_sid = partner_enz_sids.pop()
            sp = self.xba_model.species[enz_sid]
            parts_str = []
            for mr_rid, enz_stoic in sp.m_reactions.items():
                parts_str.append((f'({enz_stoic} * {self.xba_model.reactions[mr_rid].expanded_kl})/' +
                                  f'{sp.compartment}'))
            conversions_str.append(' + '.join(parts_str))

        v_str = self._to_vectors(conversions_str)
        func_code = compile('def enz_conv_mb(x): return np.array([' + ', '.join(v_str) + '])',
                            '<string>', 'exec')
        enz_conv_mb = types.FunctionType(func_code.co_consts[0], self.model_params)
        enz_conv_mb_jac = types.FunctionType(self._gradients(v_str).co_consts[0], self.model_params)
        return enz_conv_mb, enz_conv_mb_jac

    def get_gammas(self):
        # ratios between protein degradation and max protein synthesis rates
        gammas_str = []
        for sid in self.all_enz_sids:
            sp = self.xba_model.species[sid]
            if sp.p_degrad_rid is not None:
                degrad_rid = sp.p_degrad_rid
                ps_rid = sp.ps_rid
                gammas_str.append(f'({self.xba_model.reactions[degrad_rid].expanded_kl})/' +
                                  f'({self.xba_model.reactions[ps_rid].expanded_kl})')
            else:
                gammas_str.append('0')
        v_str = self._to_vectors(gammas_str)
        func_code = compile('def gammas(x): return np.array([' + ', '.join(v_str) + '])',
                            '<string>', 'exec')
        gammas = types.FunctionType(func_code.co_consts[0], self.model_params)
        gammas_jac = types.FunctionType(self._gradients(v_str).co_consts[0],
                                        self.model_params)
        return gammas, gammas_jac

    def get_growth_rate(self, x):
        # gr = gr0 * (1 - sum(gammas))
        # gr0 = 1/tau
        # tau = sum(ce/psme) = sum(ce * 1/psme)
        ce_all = np.append(x[self.var_n_metabs:], self.model_params['const_enz_conc'])
        tau = self.inv_ps_rates(x).dot(ce_all)
        gr0 = 1.0/tau
        # gr = gr0 * (1.0 - sum(gammas(x)))
        gr = gr0 * (1.0 - sum(self.p_degrads(x) * self.inv_ps_rates(x)))
        return gr

    def get_growth_rate_grad(self, x):
        # gr0_grad = d(1/tau)dx = (-dtau/dx)/tau**2 = - dtau/dx * gr0**2
        # gr_grad = gr0_grad * (1 - sum(gammas)) - gr0 * sum(gammas_jac)
        ce_all = np.append(x[self.var_n_metabs:], self.model_params['const_enz_conc'])
        tau = self.inv_ps_rates(x).dot(ce_all)
        gr0 = 1.0/tau
        dtau_dx = ce_all.dot(self.inv_ps_rates_jac(x))
        dtau_dx[self.var_n_metabs:] += self.inv_ps_rates(x)[:len(self.var_enz_sids)]
        gr0_grad = -dtau_dx * gr0**2
        deg_term = (1.0 - sum(self.p_degrads(x) * self.inv_ps_rates(x)))
        deg_term_jac = (self.p_degrads(x).reshape(-1, 1) * self.inv_ps_rates_jac(x) +
                        self.inv_ps_rates(x).reshape(-1, 1)
                        * self.p_degrads_jac(x)).sum(axis=0)
        gr_grad = gr0_grad * deg_term - gr0 * deg_term_jac
        return np.nan_to_num(gr_grad, nan=0.0)

    def get_alphas(self, x):
        # alpha_e = gr * c_e/psm_e + gamma_e
        ce_all = np.append(x[self.var_n_metabs:], self.model_params['const_enz_conc'])
        alpha0 = self.get_growth_rate(x) * self.inv_ps_rates(x)*ce_all
        deg_term = self.p_degrads(x) * self.inv_ps_rates(x)
        alphas = alpha0 + deg_term
        return alphas

    def heq_mass_balance(self, x):
        # considering syntesis costs for protein degradation
        # mass balance for species involved in metabolic reactions:
        # dMb/dt: S.dot(mrs) + deg_term + gr*(P.dot(ce) - cm) = 0
        # deg_term: P.dot(pdegs)
        # assuming protein degradation functions (pdegs) have soichiomtry of -1 for enzyme
        cm_var = x[:self.var_n_metabs]
        ce_all = np.append(x[self.var_n_metabs:], self.model_params['const_enz_conc'])
        degrad_costs = self.model_params['P_scaled_mat'].dot(self.p_degrads(x))
        metabolism = self.model_params['S_scaled_mat'].dot(self.mrs(x))
        pem = self.model_params['P_scaled_mat'].dot(ce_all) - cm_var
        dilution = self.get_growth_rate(x) * pem
        metabs_mb = metabolism + dilution + degrad_costs
        return metabs_mb

    def heq_mass_balance_jac(self, x):
        cm_var = x[:self.var_n_metabs]
        ce_all = np.append(x[self.var_n_metabs:], self.model_params['const_enz_conc'])

        metabolism_jac = self.model_params['S_scaled_mat'].dot(self.mrs_jac(x))
        degrad_costs_jac = self.model_params['P_scaled_mat'].dot(self.p_degrads_jac(x))

        # jacobian of dilution term: d/dx gr * (P.dot(ce) - cm)
        #   jac = gr_jac * (P.dot(ce) - cm) + gr * (jac(P.dot(ce)) - jac(cm))
        pem = self.model_params['P_scaled_mat'].dot(ce_all) - cm_var
        # construct jacobian of dilution term (P.dot(ce) - cm):
        #    d/dx (P.dot(ce) - cm) = d(P.dot(ce) - cm)/dx = P.dot(dce/dx) - dcm/dx
        cm_jac = -np.eye(self.var_n_metabs)
        pe_jac = self.model_params['P_scaled_mat'][:, :len(self.var_enz_sids)]
        pem_jac = np.hstack([cm_jac, pe_jac])
        dilution_jac = np.outer(pem, self.get_growth_rate_grad(x)) + self.get_growth_rate(x) * pem_jac
        metabs_mb_jac = metabolism_jac + dilution_jac + degrad_costs_jac
        return np.nan_to_num(metabs_mb_jac, nan=0.0)

    def heq_enz_conv_mb(self, x):
        return self.enz_conv_mb(x)

    def heq_enz_conv_mb_jac(self, x):
        return self.enz_conv_mb_jac(x)

    def heq_density(self, x):
        ce_var = x[self.var_n_metabs:]
        if self.model_params.get('macro', True) is True:
            density = self.model_params['weights'][self.var_n_metabs:].dot(ce_var)
        else:
            density = self.model_params['weights'].dot(x)
        density += self.model_params['const_enz_weights'].dot(
                   self.model_params['const_enz_conc'])
        return self.model_params['rho'] - density

    def heq_density_jac(self, x):
        if self.model_params.get('macro', True) is True:
            density_grad = np.zeros(len(self.model_params['weights']))
            density_grad[self.var_n_metabs:] = \
                -self.model_params['weights'][self.var_n_metabs:]
        else:
            density_grad = -self.model_params['weights']
        return density_grad
